import torch
import torch.nn as nn
import torch.nn.functional as F

class GroupNorm8(nn.Module):
    def __init__(self, num_features, eps=1e-7, affine=True):
        super().__init__()
        self.groupnorm = torch.nn.GroupNorm(num_groups=8,
                                      num_channels=num_features,
                                      eps=eps,
                                      affine=affine)
    def forward(self, x):
        return self.groupnorm(x)

class ConvNormActND(nn.Module):
    def __init__(self, dim, in_channels, out_channels, kernel_size=3, stride=1, groups=1,
                 bias=False, norm_layer=None, act_layer=nn.GELU):
        super().__init__()
        Conv = [nn.Conv1d, nn.Conv2d, nn.Conv3d][dim-1]
        Norm = norm_layer or [nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d][dim-1]
        padding = kernel_size // 2
        self.conv = Conv(in_channels, out_channels, kernel_size, stride,
                         padding=padding, groups=groups, bias=bias)
        self.norm = Norm(out_channels)
        self.act = act_layer() if act_layer else nn.Identity()

    def forward(self, x):
        return self.act(self.norm(self.conv(x)))

class MBConvND(nn.Module):
    def __init__(self, dim, in_channels, expand_ratio=6, norm_layer=None, act_layer=nn.GELU):
        super().__init__()
        mid_channels = in_channels * expand_ratio
        self.inverted_conv = ConvNormActND(dim, in_channels, mid_channels, 1, norm_layer=None, act_layer=act_layer, bias=True)
        self.depth_conv = ConvNormActND(dim, mid_channels, mid_channels, 3, groups=mid_channels, norm_layer=None, act_layer=act_layer, bias=True)
        self.point_conv = ConvNormActND(dim, mid_channels, in_channels, 1, norm_layer=norm_layer, act_layer=None, bias=False)

    def forward(self, x):
        x = self.inverted_conv(x)
        x = self.depth_conv(x)
        x = self.point_conv(x)
        return x

class ResidualBlockND(nn.Module):
    def __init__(self, main):
        super().__init__()
        self.main = main

    def forward(self, x):
        return x + self.main(x)

class LiteMLAND(nn.Module):
    def __init__(self, dim, in_channels, head_dim=32, norm_layer=None, kernel_func=nn.ReLU, eps=1e-5):
        super().__init__()
        self.eps = eps
        self.dim = dim
        heads = in_channels // head_dim
        total_dim = heads * head_dim
        Conv = [nn.Conv1d, nn.Conv2d, nn.Conv3d][dim-1]
        self.qkv = Conv(in_channels, 3 * total_dim, 1, bias=False)
        self.kernel_func = kernel_func()
        self.proj = ConvNormActND(dim, total_dim, in_channels, 1, norm_layer=norm_layer, act_layer=None)

    def forward(self, x):
        B, C = x.shape[:2]
        spatial_dims = x.shape[2:]
        N = torch.prod(torch.tensor(spatial_dims))
        qkv = self.qkv(x).reshape(B, -1, 3, N).permute(0, 3, 1, 2)
        q, k, v = qkv.unbind(-1)

        q = self.kernel_func(q)
        k = self.kernel_func(k)
        v = F.pad(v, (0, 1), mode='constant', value=1.)

        kv = torch.einsum('bnc,bnv->bcv', k, v)
        out = torch.einsum('bnc,bcv->bnv', q, kv)
        out = out[..., :-1] / (out[..., -1:] + self.eps)

        out = out.permute(0, 2, 1).reshape(B, -1, *spatial_dims)
        return self.proj(out)

class EfficientVitBlockND(nn.Module):
    def __init__(self, dim, in_channels, norm_layer=None, act_layer=nn.GELU, head_dim=32):
        super().__init__()
        self.context_module = ResidualBlockND(LiteMLAND(dim, in_channels, norm_layer=norm_layer, head_dim=head_dim))
        self.local_module = ResidualBlockND(MBConvND(dim, in_channels, norm_layer=norm_layer, act_layer=act_layer))

    def forward(self, x):
        x = self.context_module(x)
        x = self.local_module(x)
        return x

class EfficientVitLargeStageND(nn.Module):
    def __init__(self, dim, in_chs, depth, norm_layer=None, act_layer=nn.GELU, head_dim=32):
        super().__init__()
        self.blocks = nn.Sequential(*[
            EfficientVitBlockND(dim, in_chs, norm_layer=norm_layer, act_layer=act_layer, head_dim=32)
            for _ in range(depth)
        ])

    def forward(self, x):
        return self.blocks(x)



import torch
import torch.nn as nn
import einops
from .evit import GroupNorm8
from timm.models.efficientvit_mit import GELUTanh

class FactorizedConvND(nn.Module):
    def __init__(self, dim, in_chs, out_chs, kernel_size=1, bias=False):
        super().__init__()
        Conv = [nn.Conv1d, nn.Conv2d, nn.Conv3d][dim-1]
        self.dim = dim
        self.in_chs, self.out_chs = in_chs, out_chs
        g1, g2 = self._pick_groups(in_chs, out_chs)
        self.g1, self.g2 = g1, g2
        self.conv1 = Conv(in_chs, in_chs, kernel_size, groups=g1, bias=bias, padding=(kernel_size - 1)//2)
        self.conv2 = Conv(in_chs, out_chs, kernel_size, groups=g2, bias=bias, padding=(kernel_size - 1)//2)

    def _pick_groups(self, in_chs, out_chs):
        def _divisors(n):
            divs = []
            i = 1
            while i * i <= n:
                if n % i == 0:
                    divs.append(i)
                    if i != n // i:
                        divs.append(n // i)
                i += 1
            return sorted(divs)
        divs = _divisors(in_chs)
        best_g1, best_g2, best_cost = 1, in_chs, float('inf')
        for d in divs:
            g1, g2 = d, in_chs // d
            cost = (in_chs * in_chs / g1) + (in_chs * out_chs / g2)
            if cost < best_cost:
                best_cost, best_g1, best_g2 = cost, g1, g2
        return best_g1, best_g2

    def forward(self, x):
        x = self.conv1(x)
        x = einops.rearrange(x, 'b (g1 g2) ... -> b (g2 g1) ...', g1=self.g1, g2=self.g2)
        x = self.conv2(x)
        return x

class FactorizedSqueezeExciteND(nn.Module):
    def __init__(self, dim, channels, se_ratio=0.25):
        super().__init__()
        AdaptivePool = [nn.AdaptiveAvgPool1d, nn.AdaptiveAvgPool2d, nn.AdaptiveAvgPool3d][dim-1]
        self.pool = AdaptivePool(1)
        mid_chs = int(channels * se_ratio)
        self.conv_reduce = FactorizedConvND(dim, channels, mid_chs, kernel_size=1, bias=True)
        self.act = GELUTanh()
        self.conv_expand = FactorizedConvND(dim, mid_chs, channels, kernel_size=1, bias=True)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        s = self.pool(x)
        s = self.conv_reduce(s)
        s = self.act(s)
        s = self.conv_expand(s)
        s = self.sigmoid(s)
        return x * s

class FactorizedResBlockGNND(nn.Module):
    def __init__(self, dim, channels, se_ratio=0.25):
        super().__init__()
        self.conv1 = FactorizedConvND(dim, channels, channels, kernel_size=3)
        self.gn1 = GroupNorm8(channels)
        self.act1 = GELUTanh()
        self.conv2 = FactorizedConvND(dim, channels, channels, kernel_size=3)
        self.gn2 = GroupNorm8(channels)
        self.se = FactorizedSqueezeExciteND(dim, channels, se_ratio=se_ratio)

    def forward(self, x):
        skip = x
        out = self.conv1(x)
        out = self.gn1(out)
        out = self.act1(out)
        out = self.conv2(out)
        out = self.gn2(out)
        out = self.se(out)
        out = out + skip
        return out